﻿#pragma warning disable CS4014 // この呼び出しを待たないため、現在のメソッドの実行は、呼び出しが完了する前に続行します

using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Net.Http;
using System.Text;
using System.Threading.Tasks;
using System.Windows;
using Livet;
using Newtonsoft.Json;
using Newtonsoft.Json.Linq;
using Windows.Foundation;
using Windows.Media.SpeechRecognition;

namespace PpapCollage
{
    public class MainViewModel : ViewModel
    {
        private const double VideoWidth = 640;
        private const double VideoHeight = 360;
        private const double ObjectWidth = 120;
        private const double ObjectHeight = 120;

        private static readonly string s_dir = Path.GetDirectoryName(typeof(MainViewModel).Assembly.Location);

        public ObjectViewModel Pen1 { get; } = new ObjectViewModel();
        public ObjectViewModel Apple { get; } = new ObjectViewModel();
        public ObjectViewModel Pen2 { get; } = new ObjectViewModel();
        public ObjectViewModel Pineapple { get; } = new ObjectViewModel();

        private PenPineappleApplePen _ppap;
        private Dictionary<string, string[]> _searchCache;
        private bool _isRecognizing;
        private SpeechRecognizer _recognizer;
        private readonly Random _random = new Random();
        private readonly HttpClient _httpClient = new HttpClient();

        public void Initialize()
        {
            using (var reader = new ExoReader(new StreamReader(
                Path.Combine(s_dir, "ppap2.exo"),
                Encoding.GetEncoding(932)
            )))
            {
                _ppap = PenPineappleApplePen.Load(reader.ReadAll());
            }

            try
            {
                _searchCache = JsonConvert.DeserializeObject<Dictionary<string, string[]>>(
                    File.ReadAllText(Path.Combine(s_dir, "cache.json")));
            }
            catch
            {
                _searchCache = new Dictionary<string, string[]>();
            }

            _recognizer = new SpeechRecognizer();
            _recognizer.Timeouts.InitialSilenceTimeout = TimeSpan.FromSeconds(2);
            _recognizer.CompileConstraintsAsync();

            //this.Pen1.Source = new Uri("http://img1.kakaku.k-img.com/images/stationery/article/high-class_pen/p04_item03.jpg");
            //this.Pen2.Source = new Uri("http://release.co.jp/wp/wp-content/uploads/2016/02/pda-pen12bk_ma-14560317404kn8g.jpg");
            //this.Apple.Source = new Uri("https://kotobank.jp/image/dictionary/nipponica/media/81306024002557.jpg");
            //this.Pineapple.Source = new Uri("http://healing-plant.blog.so-net.ne.jp/_images/blog/_252/healing-plant/pineapple.jpg");
        }

        private string _title = "PpapCollage";
        public string Title
        {
            get
            {
                return _title;
            }
            set
            {
                if (value != _title)
                {
                    _title = value;
                    this.RaisePropertyChanged();
                }
            }
        }

        public void Update(TimeSpan position, double width, double height)
        {
            this.Update(_ppap.Pen1Animations, this.Pen1, position, width, height);
            this.Update(_ppap.AppleAnimations, this.Apple, position, width, height);
            this.Update(_ppap.Pen2Animations, this.Pen2, position, width, height);
            this.Update(_ppap.PineappleAnimations, this.Pineapple, position, width, height);

            if (!_isRecognizing)
            {
                if (!this.Pen1.IsRecognized && position >= TimeSpan.FromSeconds(18))
                {
                    this.Recognize(this.Pen1);
                }
                else if (!this.Apple.IsRecognized && position >= TimeSpan.FromSeconds(21))
                {
                    this.Recognize(this.Apple);
                }
                else if (!this.Pen2.IsRecognized && position >= TimeSpan.FromSeconds(25))
                {
                    this.Recognize(this.Pen2);
                }
                else if (!this.Pineapple.IsRecognized && position >= TimeSpan.FromSeconds(28))
                {
                    this.Recognize(this.Pineapple);
                }
            }
        }

        private void Update(IReadOnlyList<ObjectAnimation> anims, ObjectViewModel o, TimeSpan position, double width, double height)
        {
            this.Title = _isRecognizing ? "認識中" : position.ToString();

            var currentFrame = position.TotalSeconds * _ppap.Rate / _ppap.Scale;
            var anim = FindFrame(anims, (int)currentFrame);

            if (anim == null)
            {
                o.Visibility = Visibility.Collapsed;
                return;
            }

            o.Visibility = Visibility.Visible;

            var videoScale = Math.Min(width / VideoWidth, height / VideoHeight);
            var animRate = (currentFrame - anim.StartFrame) / (anim.EndFrame - anim.StartFrame);

            var objectScale = (anim.StartPosition.Scale + ((double)anim.EndPosition.Scale - anim.StartPosition.Scale) * animRate) / 100.0 * videoScale;
            o.Width = ObjectWidth * objectScale;
            o.Height = ObjectHeight * objectScale;

            o.X = (width - o.Width) / 2 + (anim.StartPosition.X + (anim.EndPosition.X - anim.StartPosition.X) * animRate) * videoScale;
            o.Y = (height - o.Height) / 2 + (anim.StartPosition.Y + (anim.EndPosition.Y - anim.StartPosition.Y) * animRate) * videoScale;
        }

        /// <summary>
        /// 指定したフレームが含まれる <see cref="ObjectAnimation"/> を二分探索
        /// </summary>
        private static ObjectAnimation FindFrame(IReadOnlyList<ObjectAnimation> anims, int frameNum)
        {
            return FindFrameCore(anims, frameNum, 0, anims.Count);
        }

        private static ObjectAnimation FindFrameCore(IReadOnlyList<ObjectAnimation> anims, int frameNum, int startIndex, int endIndex)
        {
            if (startIndex == endIndex) return null;

            ObjectAnimation a;

            if (endIndex - startIndex == 1)
            {
                a = anims[startIndex];
                return frameNum >= a.StartFrame && frameNum <= a.EndFrame
                    ? a : null;
            }

            var i = startIndex + (endIndex - startIndex) / 2;
            a = anims[i];

            if (frameNum >= a.StartFrame)
            {
                if (frameNum <= a.EndFrame) return a;
                return FindFrameCore(anims, frameNum, i + 1, endIndex);
            }

            return FindFrameCore(anims, frameNum, startIndex, i);
        }

        private void Recognize(ObjectViewModel o)
        {
            _isRecognizing = true;

            _recognizer.RecognizeAsync().Completed += (info, status) =>
            {
                o.IsRecognized = true;
                _isRecognizing = false;

                var text = "無";

                if (status == AsyncStatus.Completed)
                {
                    var s = info.GetResults().Text;
                    if (!string.IsNullOrEmpty(s)) text = s;
                }

                this.SetImage(o, text);
            };
        }

        private async void SetImage(ObjectViewModel o, string text)
        {
            string[] images;

            if (!_searchCache.TryGetValue(text, out images))
            {
                var uri = "https://www.googleapis.com/customsearch/v1?key=AIzaSyAvm_4WZcpQIlB0PY6yx6xMxuwz7gxQZ4A&cx=015001858452408684756:cdvklqpe4yk&searchType=image&q="
                    + Uri.EscapeDataString(text);
                var res = await _httpClient.GetStringAsync(new Uri(uri));
                var j = JObject.Parse(res)["items"];

                if (j != null)
                {
                    images = JObject.Parse(res)["items"]
                        .Select(x => x["link"])
                        .Where(x => x != null)
                        .Select(x => (string)x)
                        .ToArray();
                }

                _searchCache[text] = images;

                Task.Run(() =>
                {
                    try
                    {
                        File.WriteAllText(Path.Combine(s_dir, "cache.json"), JsonConvert.SerializeObject(_searchCache));
                    }
                    catch { }
                });
            }

            if (images.Length > 0)
            {
                App.Current.Dispatcher.InvokeAsync(() =>
                {
                    o.Source = new Uri(images[_random.Next(images.Length)]);
                });
            }
        }
    }
}
